package Test::LineMap;

use strict;
use warnings;

use Time::HiRes qw(usleep);

#line 40 "one-file.pm"

usleep(50000);

#line 30 "other-file.pm"

usleep(50000);

#line 20 "one-file.pm"

usleep(50000);

#line 40 "other-file.pm"

1;
