#ifndef _DEVEL_STATPROFILER_RAND
#define _DEVEL_STATPROFILER_RAND

namespace devel {
    namespace statprofiler {
        unsigned int rand_seed();
        void rand(unsigned int *seed);
    }
}

#endif
